import http
import logging
import unittest
import unittest.mock

from websockets.datastructures import Headers
from websockets.exceptions import (
    InvalidHeader,
    InvalidOrigin,
    InvalidUpgrade,
    NegotiationError,
)
from websockets.frames import OP_TEXT, Frame
from websockets.http11 import Request, Response
from websockets.protocol import CONNECTING, OPEN
from websockets.server import *

from .extensions.utils import (
    OpExtension,
    Rsv2Extension,
    ServerOpExtensionFactory,
    ServerRsv2ExtensionFactory,
)
from .test_utils import ACCEPT, KEY
from .utils import DATE, DeprecationTestCase


class ConnectTests(unittest.TestCase):
    def test_receive_connect(self):
        server = ServerProtocol()
        server.receive_data(
            (
                f"GET /test HTTP/1.1\r\n"
                f"Host: example.com\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Key: {KEY}\r\n"
                f"Sec-WebSocket-Version: 13\r\n"
                f"\r\n"
            ).encode(),
        )
        [request] = server.events_received()
        self.assertIsInstance(request, Request)
        self.assertEqual(server.data_to_send(), [])
        self.assertFalse(server.close_expected())

    def test_connect_request(self):
        server = ServerProtocol()
        server.receive_data(
            (
                f"GET /test HTTP/1.1\r\n"
                f"Host: example.com\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Key: {KEY}\r\n"
                f"Sec-WebSocket-Version: 13\r\n"
                f"\r\n"
            ).encode(),
        )
        [request] = server.events_received()
        self.assertEqual(request.path, "/test")
        self.assertEqual(
            request.headers,
            Headers(
                {
                    "Host": "example.com",
                    "Upgrade": "websocket",
                    "Connection": "Upgrade",
                    "Sec-WebSocket-Key": KEY,
                    "Sec-WebSocket-Version": "13",
                }
            ),
        )

    def test_no_request(self):
        server = ServerProtocol()
        server.receive_eof()
        self.assertEqual(server.events_received(), [])

    def test_partial_request(self):
        server = ServerProtocol()
        server.receive_data(b"GET /test HTTP/1.1\r\n")
        server.receive_eof()
        self.assertEqual(server.events_received(), [])

    def test_random_request(self):
        server = ServerProtocol()
        server.receive_data(b"HELO relay.invalid\r\n")
        server.receive_data(b"MAIL FROM: <alice@invalid>\r\n")
        server.receive_data(b"RCPT TO: <bob@invalid>\r\n")
        self.assertEqual(server.events_received(), [])


class AcceptRejectTests(unittest.TestCase):
    def make_request(self):
        return Request(
            path="/test",
            headers=Headers(
                {
                    "Host": "example.com",
                    "Upgrade": "websocket",
                    "Connection": "Upgrade",
                    "Sec-WebSocket-Key": KEY,
                    "Sec-WebSocket-Version": "13",
                }
            ),
        )

    def test_send_accept(self):
        server = ServerProtocol()
        with unittest.mock.patch("email.utils.formatdate", return_value=DATE):
            response = server.accept(self.make_request())
        self.assertIsInstance(response, Response)
        server.send_response(response)
        self.assertEqual(
            server.data_to_send(),
            [
                f"HTTP/1.1 101 Switching Protocols\r\n"
                f"Date: {DATE}\r\n"
                f"Upgrade: websocket\r\n"
                f"Connection: Upgrade\r\n"
                f"Sec-WebSocket-Accept: {ACCEPT}\r\n"
                f"\r\n".encode()
            ],
        )
        self.assertFalse(server.close_expected())
        self.assertEqual(server.state, OPEN)

    def test_send_reject(self):
        server = ServerProtocol()
        with unittest.mock.patch("email.utils.formatdate", return_value=DATE):
            response = server.reject(http.HTTPStatus.NOT_FOUND, "Sorry folks.\n")
        self.assertIsInstance(response, Response)
        server.send_response(response)
        self.assertEqual(
            server.data_to_send(),
            [
                f"HTTP/1.1 404 Not Found\r\n"
                f"Date: {DATE}\r\n"
                f"Connection: close\r\n"
                f"Content-Length: 13\r\n"
                f"Content-Type: text/plain; charset=utf-8\r\n"
                f"\r\n"
                f"Sorry folks.\n".encode(),
                b"",
            ],
        )
        self.assertTrue(server.close_expected())
        self.assertEqual(server.state, CONNECTING)

    def test_accept_response(self):
        server = ServerProtocol()
        with unittest.mock.patch("email.utils.formatdate", return_value=DATE):
            response = server.accept(self.make_request())
        self.assertIsInstance(response, Response)
        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.reason_phrase, "Switching Protocols")
        self.assertEqual(
            response.headers,
            Headers(
                {
                    "Date": DATE,
                    "Upgrade": "websocket",
                    "Connection": "Upgrade",
                    "Sec-WebSocket-Accept": ACCEPT,
                }
            ),
        )
        self.assertIsNone(response.body)

    def test_reject_response(self):
        server = ServerProtocol()
        with unittest.mock.patch("email.utils.formatdate", return_value=DATE):
            response = server.reject(http.HTTPStatus.NOT_FOUND, "Sorry folks.\n")
        self.assertIsInstance(response, Response)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.reason_phrase, "Not Found")
        self.assertEqual(
            response.headers,
            Headers(
                {
                    "Date": DATE,
                    "Connection": "close",
                    "Content-Length": "13",
                    "Content-Type": "text/plain; charset=utf-8",
                }
            ),
        )
        self.assertEqual(response.body, b"Sorry folks.\n")

    def test_reject_response_supports_int_status(self):
        server = ServerProtocol()
        response = server.reject(404, "Sorry folks.\n")
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.reason_phrase, "Not Found")

    def test_basic(self):
        server = ServerProtocol()
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)

    def test_unexpected_exception(self):
        server = ServerProtocol()
        request = self.make_request()
        with unittest.mock.patch(
            "websockets.server.ServerProtocol.process_request",
            side_effect=Exception("BOOM"),
        ):
            response = server.accept(request)

        self.assertEqual(response.status_code, 500)
        with self.assertRaises(Exception) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "BOOM")

    def test_missing_connection(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Connection"]
        response = server.accept(request)

        self.assertEqual(response.status_code, 426)
        self.assertEqual(response.headers["Upgrade"], "websocket")
        with self.assertRaises(InvalidUpgrade) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "missing Connection header")

    def test_invalid_connection(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Connection"]
        request.headers["Connection"] = "close"
        response = server.accept(request)

        self.assertEqual(response.status_code, 426)
        self.assertEqual(response.headers["Upgrade"], "websocket")
        with self.assertRaises(InvalidUpgrade) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "invalid Connection header: close")

    def test_missing_upgrade(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Upgrade"]
        response = server.accept(request)

        self.assertEqual(response.status_code, 426)
        self.assertEqual(response.headers["Upgrade"], "websocket")
        with self.assertRaises(InvalidUpgrade) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "missing Upgrade header")

    def test_invalid_upgrade(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Upgrade"]
        request.headers["Upgrade"] = "h2c"
        response = server.accept(request)

        self.assertEqual(response.status_code, 426)
        self.assertEqual(response.headers["Upgrade"], "websocket")
        with self.assertRaises(InvalidUpgrade) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "invalid Upgrade header: h2c")

    def test_missing_key(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Sec-WebSocket-Key"]
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "missing Sec-WebSocket-Key header")

    def test_multiple_key(self):
        server = ServerProtocol()
        request = self.make_request()
        request.headers["Sec-WebSocket-Key"] = KEY
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception),
            "invalid Sec-WebSocket-Key header: "
            "more than one Sec-WebSocket-Key header found",
        )

    def test_invalid_key(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Sec-WebSocket-Key"]
        request.headers["Sec-WebSocket-Key"] = "not Base64 data!"
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception), "invalid Sec-WebSocket-Key header: not Base64 data!"
        )

    def test_truncated_key(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Sec-WebSocket-Key"]
        request.headers["Sec-WebSocket-Key"] = KEY[
            :16
        ]  # 12 bytes instead of 16, Base64-encoded
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception), f"invalid Sec-WebSocket-Key header: {KEY[:16]}"
        )

    def test_missing_version(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Sec-WebSocket-Version"]
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "missing Sec-WebSocket-Version header")

    def test_multiple_version(self):
        server = ServerProtocol()
        request = self.make_request()
        request.headers["Sec-WebSocket-Version"] = "11"
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception),
            "invalid Sec-WebSocket-Version header: "
            "more than one Sec-WebSocket-Version header found",
        )

    def test_invalid_version(self):
        server = ServerProtocol()
        request = self.make_request()
        del request.headers["Sec-WebSocket-Version"]
        request.headers["Sec-WebSocket-Version"] = "11"
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception), "invalid Sec-WebSocket-Version header: 11"
        )

    def test_no_origin(self):
        server = ServerProtocol(origins=["https://example.com"])
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 403)
        with self.assertRaises(InvalidOrigin) as raised:
            raise server.handshake_exc
        self.assertEqual(str(raised.exception), "missing Origin header")

    def test_origin(self):
        server = ServerProtocol(origins=["https://example.com"])
        request = self.make_request()
        request.headers["Origin"] = "https://example.com"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(server.origin, "https://example.com")

    def test_unexpected_origin(self):
        server = ServerProtocol(origins=["https://example.com"])
        request = self.make_request()
        request.headers["Origin"] = "https://other.example.com"
        response = server.accept(request)

        self.assertEqual(response.status_code, 403)
        with self.assertRaises(InvalidOrigin) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception), "invalid Origin header: https://other.example.com"
        )

    def test_multiple_origin(self):
        server = ServerProtocol(
            origins=["https://example.com", "https://other.example.com"]
        )
        request = self.make_request()
        request.headers["Origin"] = "https://example.com"
        request.headers["Origin"] = "https://other.example.com"
        response = server.accept(request)

        # This is prohibited by the HTTP specification, so the return code is
        # 400 Bad Request rather than 403 Forbidden.
        self.assertEqual(response.status_code, 400)
        with self.assertRaises(InvalidHeader) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception),
            "invalid Origin header: more than one Origin header found",
        )

    def test_supported_origin(self):
        server = ServerProtocol(
            origins=["https://example.com", "https://other.example.com"]
        )
        request = self.make_request()
        request.headers["Origin"] = "https://other.example.com"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(server.origin, "https://other.example.com")

    def test_unsupported_origin(self):
        server = ServerProtocol(
            origins=["https://example.com", "https://other.example.com"]
        )
        request = self.make_request()
        request.headers["Origin"] = "https://original.example.com"
        response = server.accept(request)

        self.assertEqual(response.status_code, 403)
        with self.assertRaises(InvalidOrigin) as raised:
            raise server.handshake_exc
        self.assertEqual(
            str(raised.exception), "invalid Origin header: https://original.example.com"
        )

    def test_no_origin_accepted(self):
        server = ServerProtocol(origins=[None])
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertIsNone(server.origin)

    def test_no_extensions(self):
        server = ServerProtocol()
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Extensions", response.headers)
        self.assertEqual(server.extensions, [])

    def test_no_extension(self):
        server = ServerProtocol(extensions=[ServerOpExtensionFactory()])
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Extensions", response.headers)
        self.assertEqual(server.extensions, [])

    def test_extension(self):
        server = ServerProtocol(extensions=[ServerOpExtensionFactory()])
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Extensions"], "x-op; op")
        self.assertEqual(server.extensions, [OpExtension()])

    def test_unexpected_extension(self):
        server = ServerProtocol()
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Extensions", response.headers)
        self.assertEqual(server.extensions, [])

    def test_unsupported_extension(self):
        server = ServerProtocol(extensions=[ServerRsv2ExtensionFactory()])
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Extensions", response.headers)
        self.assertEqual(server.extensions, [])

    def test_supported_extension_parameters(self):
        server = ServerProtocol(extensions=[ServerOpExtensionFactory("this")])
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op=this"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Extensions"], "x-op; op=this")
        self.assertEqual(server.extensions, [OpExtension("this")])

    def test_unsupported_extension_parameters(self):
        server = ServerProtocol(extensions=[ServerOpExtensionFactory("this")])
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op=that"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Extensions", response.headers)
        self.assertEqual(server.extensions, [])

    def test_multiple_supported_extension_parameters(self):
        server = ServerProtocol(
            extensions=[
                ServerOpExtensionFactory("this"),
                ServerOpExtensionFactory("that"),
            ]
        )
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op=that"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Extensions"], "x-op; op=that")
        self.assertEqual(server.extensions, [OpExtension("that")])

    def test_multiple_extensions(self):
        server = ServerProtocol(
            extensions=[ServerOpExtensionFactory(), ServerRsv2ExtensionFactory()]
        )
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op"
        request.headers["Sec-WebSocket-Extensions"] = "x-rsv2"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(
            response.headers["Sec-WebSocket-Extensions"], "x-op; op, x-rsv2"
        )
        self.assertEqual(server.extensions, [OpExtension(), Rsv2Extension()])

    def test_multiple_extensions_order(self):
        server = ServerProtocol(
            extensions=[ServerOpExtensionFactory(), ServerRsv2ExtensionFactory()]
        )
        request = self.make_request()
        request.headers["Sec-WebSocket-Extensions"] = "x-rsv2"
        request.headers["Sec-WebSocket-Extensions"] = "x-op; op"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(
            response.headers["Sec-WebSocket-Extensions"], "x-rsv2, x-op; op"
        )
        self.assertEqual(server.extensions, [Rsv2Extension(), OpExtension()])

    def test_no_subprotocols(self):
        server = ServerProtocol()
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Protocol", response.headers)
        self.assertIsNone(server.subprotocol)

    def test_no_subprotocol(self):
        server = ServerProtocol(subprotocols=["chat"])
        request = self.make_request()
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaisesRegex(
            NegotiationError,
            r"missing subprotocol",
        ):
            raise server.handshake_exc

    def test_subprotocol(self):
        server = ServerProtocol(subprotocols=["chat"])
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "chat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Protocol"], "chat")
        self.assertEqual(server.subprotocol, "chat")

    def test_unexpected_subprotocol(self):
        server = ServerProtocol()
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "chat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Protocol", response.headers)
        self.assertIsNone(server.subprotocol)

    def test_multiple_subprotocols(self):
        server = ServerProtocol(subprotocols=["superchat", "chat"])
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "chat"
        request.headers["Sec-WebSocket-Protocol"] = "superchat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Protocol"], "superchat")
        self.assertEqual(server.subprotocol, "superchat")

    def test_supported_subprotocol(self):
        server = ServerProtocol(subprotocols=["superchat", "chat"])
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "chat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Protocol"], "chat")
        self.assertEqual(server.subprotocol, "chat")

    def test_unsupported_subprotocol(self):
        server = ServerProtocol(subprotocols=["superchat", "chat"])
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "otherchat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 400)
        with self.assertRaisesRegex(
            NegotiationError,
            r"invalid subprotocol; expected one of superchat, chat",
        ):
            raise server.handshake_exc

    @staticmethod
    def optional_chat(protocol, subprotocols):
        if "chat" in subprotocols:
            return "chat"

    def test_select_subprotocol(self):
        server = ServerProtocol(select_subprotocol=self.optional_chat)
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "chat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertEqual(response.headers["Sec-WebSocket-Protocol"], "chat")
        self.assertEqual(server.subprotocol, "chat")

    def test_select_no_subprotocol(self):
        server = ServerProtocol(select_subprotocol=self.optional_chat)
        request = self.make_request()
        request.headers["Sec-WebSocket-Protocol"] = "otherchat"
        response = server.accept(request)

        self.assertEqual(response.status_code, 101)
        self.assertNotIn("Sec-WebSocket-Protocol", response.headers)
        self.assertIsNone(server.subprotocol)


class MiscTests(unittest.TestCase):
    def test_bypass_handshake(self):
        server = ServerProtocol(state=OPEN)
        server.receive_data(b"\x81\x86\x00\x00\x00\x00Hello!")
        [frame] = server.events_received()
        self.assertEqual(frame, Frame(OP_TEXT, b"Hello!"))

    def test_custom_logger(self):
        logger = logging.getLogger("test")
        with self.assertLogs("test", logging.DEBUG) as logs:
            ServerProtocol(logger=logger)
        self.assertEqual(len(logs.records), 1)


class BackwardsCompatibilityTests(DeprecationTestCase):
    def test_server_connection_class(self):
        with self.assertDeprecationWarning(
            "ServerConnection was renamed to ServerProtocol"
        ):
            from websockets.server import ServerConnection

            server = ServerConnection()

        self.assertIsInstance(server, ServerProtocol)



# SPDX-License-Identifier: MIT

# Keep this file SHORT, until Black can handle it.
import pytest

import attr


class TestPatternMatching:
    """
    Pattern matching syntax test cases.
    """

    @pytest.mark.parametrize("dec", [attr.s, attr.define, attr.frozen])
    def test_simple_match_case(self, dec):
        """
        Simple match case statement works as expected with all class
        decorators.
        """

        @dec
        class C(object):
            a = attr.ib()

        assert ("a",) == C.__match_args__

        matched = False
        c = C(a=1)
        match c:
            case C(a):
                matched = True

        assert matched
        assert 1 == a

    def test_explicit_match_args(self):
        """
        Does not overwrite a manually set empty __match_args__.
        """

        ma = ()

        @attr.define
        class C:
            a = attr.field()
            __match_args__ = ma

        c = C(a=1)

        msg = r"C\(\) accepts 0 positional sub-patterns \(1 given\)"
        with pytest.raises(TypeError, match=msg):
            match c:
                case C(_):
                    pass

    def test_match_args_kw_only(self):
        """
        kw_only classes don't generate __match_args__.
        kw_only fields are not included in __match_args__.
        """

        @attr.define
        class C:
            a = attr.field(kw_only=True)
            b = attr.field()

        assert ("b",) == C.__match_args__

        c = C(a=1, b=1)
        msg = r"C\(\) accepts 1 positional sub-pattern \(2 given\)"
        with pytest.raises(TypeError, match=msg):
            match c:
                case C(a, b):
                    pass

        found = False
        match c:
            case C(b, a=a):
                found = True

        assert found

        @attr.define(kw_only=True)
        class C:
            a = attr.field()
            b = attr.field()

        c = C(a=1, b=1)
        msg = r"C\(\) accepts 0 positional sub-patterns \(2 given\)"
        with pytest.raises(TypeError, match=msg):
            match c:
                case C(a, b):
                    pass

        found = False
        match c:
            case C(a=a, b=b):
                found = True

        assert found
        assert (1, 1) == (a, b)



import io
import os
import sys

import pyperf

sys.path[0:0] = [os.path.join(os.path.dirname(__file__), "..")]
import html5lib  # noqa: E402


def bench_html5lib(fh):
    fh.seek(0)
    html5lib.parse(fh, treebuilder="etree", useChardet=False)


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


BENCHMARKS = {}
for root, dirs, files in os.walk(os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "wpt")):
    for f in files:
        if f.endswith(".html"):
            BENCHMARKS[f[: -len(".html")]] = os.path.join(root, f)


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "Run parser benchmarks from WPT"
    runner.argparser.add_argument("benchmark", nargs="?", choices=sorted(BENCHMARKS))

    args = runner.parse_args()
    if args.benchmark:
        benchmarks = (args.benchmark,)
    else:
        benchmarks = sorted(BENCHMARKS)

    for bench in benchmarks:
        name = "wpt_%s" % bench
        path = BENCHMARKS[bench]
        with open(path, "rb") as fh:
            fh2 = io.BytesIO(fh.read())

        runner.bench_func(name, bench_html5lib, fh2)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

# This file contains logging functionality for mach. It essentially provides
# support for a structured logging framework built on top of Python's built-in
# logging framework.

import codecs
import json
import logging
import os
import sys
import time

import blessed
import six
from mozbuild.buildversion import mozilla_build_version
from packaging.version import Version

IS_WINDOWS = sys.platform.startswith("win")

if IS_WINDOWS:
    import msvcrt
    from ctypes import byref, windll
    from ctypes.wintypes import DWORD

    ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004

    def enable_virtual_terminal_processing(file_descriptor):
        handle = msvcrt.get_osfhandle(file_descriptor)
        try:
            mode = DWORD()
            windll.kernel32.GetConsoleMode(handle, byref(mode))
            mode.value |= ENABLE_VIRTUAL_TERMINAL_PROCESSING
            windll.kernel32.SetConsoleMode(handle, mode.value)
        except Exception as e:
            raise e


def enable_blessed():
    # Only Windows has issues with enabling blessed
    # and interpreting ANSI escape sequences
    if not IS_WINDOWS:
        return True

    if os.environ.get("NO_ANSI"):
        return False

    if not os.environ.get("MOZILLABUILD"):
        return False

    # MozillaBuild 4.0.2 is the first Release that supports
    # ANSI escape sequences, so if we're greater than that
    # version, we can enable them (via Blessed).
    return mozilla_build_version() >= Version("4.0.2")


# stdout and stderr may not necessarily be set up to write Unicode output, so
# reconfigure them if necessary.
def _wrap_stdstream(fh):
    if fh in (sys.stderr, sys.stdout):
        encoding = sys.getdefaultencoding()
        encoding = "utf-8" if encoding in ("ascii", "charmap") else encoding
        if six.PY2:
            return codecs.getwriter(encoding)(fh, errors="replace")
        else:
            return codecs.getwriter(encoding)(fh.buffer, errors="replace")
    else:
        return fh


def format_seconds(total):
    """Format number of seconds to MM:SS.DD form."""

    minutes, seconds = divmod(total, 60)

    return "%2d:%05.2f" % (minutes, seconds)


def format_level(level, terminal=None):
    levels = {
        logging.NOTSET: ("N", "bright_white"),
        logging.DEBUG: ("D", "blue"),
        logging.INFO: (None, None),
        logging.WARNING: ("W", "yellow"),
        logging.ERROR: ("E", "red"),
        logging.CRITICAL: ("C", "black_on_red"),
    }
    try:
        s, color = levels[level]
    except KeyError:
        raise ValueError(f"unsupported log level: {level}")
    if s is None:
        return ""
    colorfunc = getattr(terminal, color) if terminal else lambda x: x
    return colorfunc(s) + " "


class ConvertToStructuredFilter(logging.Filter):
    """Filter that converts unstructured records into structured ones."""

    def filter(self, record):
        if hasattr(record, "action") and hasattr(record, "params"):
            return True

        record.action = "unstructured"
        record.params = {"msg": record.getMessage()}
        record.msg = "{msg}"

        return True


class StructuredJSONFormatter(logging.Formatter):
    """Log formatter that writes a structured JSON entry."""

    def format(self, record):
        action = getattr(record, "action", "UNKNOWN")
        params = getattr(record, "params", {})

        return json.dumps([record.created, action, params])


class StructuredHumanFormatter(logging.Formatter):
    """Log formatter that writes structured messages for humans.

    It is important that this formatter never be added to a logger that
    produces unstructured/classic log messages. If it is, the call to format()
    could fail because the string could contain things (like JSON) that look
    like formatting character sequences.

    Because of this limitation, format() will fail with a KeyError if an
    unstructured record is passed or if the structured message is malformed.
    """

    def __init__(
        self, start_time, write_interval=False, write_times=True, write_level=None
    ):
        self.start_time = start_time
        self.write_interval = write_interval
        self.write_times = write_times
        # Default to the same value as `write_times` if `write_level` is unset.
        self.write_level = write_level if write_level is not None else write_times
        self.last_time = None

    def format(self, record):
        formatted_msg = record.msg.format(**getattr(record, "params", {}))

        elapsed_time = (
            format_seconds(self._time(record)) + " " if self.write_times else ""
        )

        level = format_level(record.levelno) if self.write_level else ""

        rv = elapsed_time + level + formatted_msg
        formatted_stack_trace_result = formatted_stack_trace(record, self)

        if formatted_stack_trace_result != "":
            stack_trace = "\n" + elapsed_time + formatted_stack_trace_result
            rv += stack_trace.replace("\n", f"\n{elapsed_time}")

        return rv

    def _time(self, record):
        t = record.created - self.start_time

        if self.write_interval and self.last_time is not None:
            t = record.created - self.last_time

        self.last_time = record.created

        return t


class StructuredTerminalFormatter(StructuredHumanFormatter):
    """Log formatter for structured messages writing to a terminal."""

    def set_terminal(self, terminal):
        self.terminal = terminal
        self._sgr0 = terminal.normal if terminal else ""

    def format(self, record):
        formatted_msg = record.msg.format(**getattr(record, "params", {}))
        elapsed_time = (
            self.terminal.blue(format_seconds(self._time(record))) + " "
            if self.write_times
            else ""
        )

        level = format_level(record.levelno, self.terminal) if self.write_level else ""

        rv = elapsed_time + level + self._colorize(formatted_msg) + self._sgr0
        formatted_stack_trace_result = formatted_stack_trace(record, self)

        if formatted_stack_trace_result != "":
            stack_trace = "\n" + elapsed_time + formatted_stack_trace_result
            rv += stack_trace.replace("\n", f"\n{elapsed_time}")

        # Some processes (notably Clang) don't reset terminal attributes after
        # printing newlines. This can lead to terminal attributes getting in a
        # wonky state. Work around this by sending the sgr0 sequence after every
        # line to reset all attributes. For programs that rely on the next line
        # inheriting the same attributes, this will prevent that from happening.
        # But that's better than "corrupting" the terminal.
        return rv + self._sgr0

    def _colorize(self, s):
        if not self.terminal:
            return s

        result = s

        reftest = s.startswith("REFTEST ")
        if reftest:
            s = s[8:]

        if s.startswith("TEST-PASS"):
            result = self.terminal.green(s[0:9]) + s[9:]
        elif s.startswith("TEST-UNEXPECTED"):
            result = self.terminal.red(s[0:20]) + s[20:]
        elif s.startswith("TEST-START"):
            result = self.terminal.yellow(s[0:10]) + s[10:]
        elif s.startswith("TEST-INFO"):
            result = self.terminal.yellow(s[0:9]) + s[9:]

        if reftest:
            result = "REFTEST " + result

        return result


def formatted_stack_trace(record, formatter):
    """
    Formatting behavior here intended to mimic a portion of the
    standard library's logging.Formatter::format function
    """
    rv = ""

    if record.exc_info:
        # Cache the traceback text to avoid converting it multiple times
        # (it's constant anyway)
        if not record.exc_text:
            record.exc_text = formatter.formatException(record.exc_info)
    if record.exc_text:
        rv = record.exc_text
    if record.stack_info:
        if rv[-1:] != "\n":
            rv = rv + "\n"
        rv = rv + formatter.formatStack(record.stack_info)

    return rv


class LoggingManager(object):
    """Holds and controls global logging state.

    An application should instantiate one of these and configure it as needed.

    This class provides a mechanism to configure the output of logging data
    both from mach and from the overall logging system (e.g. from other
    modules).
    """

    def __init__(self):
        self.start_time = time.time()

        self.json_handlers = []
        self.terminal_handler = None
        self.terminal_formatter = None

        self.root_logger = logging.getLogger()
        self.root_logger.setLevel(logging.DEBUG)

        # Installing NullHandler on the root logger ensures that *all* log
        # messages have at least one handler. This prevents Python from
        # complaining about "no handlers could be found for logger XXX."
        self.root_logger.addHandler(logging.NullHandler())

        mach_logger = logging.getLogger("mach")
        mach_logger.setLevel(logging.DEBUG)

        self.structured_filter = ConvertToStructuredFilter()

        self.structured_loggers = [mach_logger]

        self._terminal = None

    def create_terminal(self):
        if enable_blessed():
            # Sometimes blessed fails to set up the terminal, in that case, silently fail.
            try:
                terminal = blessed.Terminal(stream=_wrap_stdstream(sys.stdout))

                if terminal.is_a_tty:
                    self._terminal = terminal
            except Exception:
                pass

    @property
    def terminal(self):
        return self._terminal

    def add_json_handler(self, fh):
        """Enable JSON logging on the specified file object."""

        # Configure the consumer of structured messages.
        handler = logging.StreamHandler(stream=fh)
        handler.setFormatter(StructuredJSONFormatter())
        handler.setLevel(logging.DEBUG)

        # And hook it up.
        for logger in self.structured_loggers:
            logger.addHandler(handler)

        self.json_handlers.append(handler)

    def add_terminal_logging(
        self, fh=sys.stdout, level=logging.INFO, write_interval=False, write_times=True
    ):
        """Enable logging to the terminal."""
        self.create_terminal()

        if IS_WINDOWS:
            try:
                # fileno() can raise in some cases, like unit tests.
                # so we can try to enable this but if we fail it's fine
                enable_virtual_terminal_processing(sys.stdout.fileno())
                enable_virtual_terminal_processing(sys.stderr.fileno())
            except Exception:
                pass

        fh = _wrap_stdstream(fh)
        formatter = StructuredHumanFormatter(
            self.start_time, write_interval=write_interval, write_times=write_times
        )

        if self.terminal:
            formatter = StructuredTerminalFormatter(
                self.start_time, write_interval=write_interval, write_times=write_times
            )
            formatter.set_terminal(self.terminal)

        handler = logging.StreamHandler(stream=fh)
        handler.setFormatter(formatter)
        handler.setLevel(level)

        for logger in self.structured_loggers:
            logger.addHandler(handler)

        self.terminal_handler = handler
        self.terminal_formatter = formatter

    def replace_terminal_handler(self, handler):
        """Replace the installed terminal handler.

        Returns the old handler or None if none was configured.
        If the new handler is None, removes any existing handler and disables
        logging to the terminal.
        """
        old = self.terminal_handler

        if old:
            for logger in self.structured_loggers:
                logger.removeHandler(old)

        if handler:
            for logger in self.structured_loggers:
                logger.addHandler(handler)

        self.terminal_handler = handler

        return old

    def enable_unstructured(self):
        """Enable logging of unstructured messages."""
        if self.terminal_handler:
            self.terminal_handler.addFilter(self.structured_filter)
            self.root_logger.addHandler(self.terminal_handler)

    def disable_unstructured(self):
        """Disable logging of unstructured messages."""
        if self.terminal_handler:
            self.terminal_handler.removeFilter(self.structured_filter)
            self.root_logger.removeHandler(self.terminal_handler)

    def register_structured_logger(self, logger, terminal=True, json=True):
        """Register a structured logger.

        This needs to be called for all structured loggers that don't chain up
        to the mach logger in order for their output to be captured.
        """
        self.structured_loggers.append(logger)

        if terminal and self.terminal_handler:
            logger.addHandler(self.terminal_handler)

        if json:
            for handler in self.json_handlers:
                logger.addHandler(handler)

    def enable_all_structured_loggers(self, terminal=True, json=True):
        """Enable logging of all structured messages from all loggers.

        ``terminal`` and ``json`` determine which log handlers to operate
        on. By default, all known handlers are operated on.
        """

        # Glean makes logs that we're not interested in, so we squelch them.
        logging.getLogger("glean").setLevel(logging.CRITICAL)

        # Remove current handlers from all loggers so we don't double
        # register handlers.
        for logger in self.root_logger.manager.loggerDict.values():
            # Some entries might be logging.PlaceHolder.
            if not isinstance(logger, logging.Logger):
                continue

            if terminal:
                logger.removeHandler(self.terminal_handler)

            if json:
                for handler in self.json_handlers:
                    logger.removeHandler(handler)

        # Wipe out existing registered structured loggers since they
        # all propagate to root logger.
        self.structured_loggers = []
        self.register_structured_logger(self.root_logger, terminal=terminal, json=json)



try:
    from ._version import version as __version__
except ImportError:
    # broken installation, we don't even try
    # unknown only works because we do poor mans version compare
    __version__ = "unknown"

__all__ = [
    "PluginManager",
    "PluginValidationError",
    "HookCallError",
    "HookspecMarker",
    "HookimplMarker",
]

from ._manager import PluginManager, PluginValidationError
from ._callers import HookCallError
from ._hooks import HookspecMarker, HookimplMarker



from __future__ import absolute_import, division, unicode_literals

import io

from six import unichr, text_type

from html5lib._tokenizer import HTMLTokenizer
from html5lib.constants import tokenTypes


def ignore_parse_errors(toks):
    for tok in toks:
        if tok['type'] != tokenTypes['ParseError']:
            yield tok


def test_maintain_attribute_order():
    # generate loads to maximize the chance a hash-based mutation will occur
    attrs = [(unichr(x), text_type(i)) for i, x in enumerate(range(ord('a'), ord('z')))]
    stream = io.StringIO("<span " + " ".join("%s='%s'" % (x, i) for x, i in attrs) + ">")

    toks = HTMLTokenizer(stream)
    out = list(ignore_parse_errors(toks))

    assert len(out) == 1
    assert out[0]['type'] == tokenTypes['StartTag']

    attrs_tok = out[0]['data']
    assert len(attrs_tok) == len(attrs)

    for (in_name, in_value), (out_name, out_value) in zip(attrs, attrs_tok.items()):
        assert in_name == out_name
        assert in_value == out_value


def test_duplicate_attribute():
    stream = io.StringIO("<span a=1 a=2 a=3>")

    toks = HTMLTokenizer(stream)
    out = list(ignore_parse_errors(toks))

    assert len(out) == 1
    assert out[0]['type'] == tokenTypes['StartTag']

    attrs_tok = out[0]['data']
    assert len(attrs_tok) == 1
    assert list(attrs_tok.items()) == [('a', '1')]


def test_maintain_duplicate_attribute_order():
    # generate loads to maximize the chance a hash-based mutation will occur
    attrs = [(unichr(x), text_type(i)) for i, x in enumerate(range(ord('a'), ord('z')))]
    stream = io.StringIO("<span " + " ".join("%s='%s'" % (x, i) for x, i in attrs) + " a=100>")

    toks = HTMLTokenizer(stream)
    out = list(ignore_parse_errors(toks))

    assert len(out) == 1
    assert out[0]['type'] == tokenTypes['StartTag']

    attrs_tok = out[0]['data']
    assert len(attrs_tok) == len(attrs)

    for (in_name, in_value), (out_name, out_value) in zip(attrs, attrs_tok.items()):
        assert in_name == out_name
        assert in_value == out_value



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import copy

import requests

from marionette_harness import MarionetteTestCase


class TestCommandLineArguments(MarionetteTestCase):
    def setUp(self):
        super(TestCommandLineArguments, self).setUp()

        self.orig_arguments = copy.copy(self.marionette.instance.app_args)

    def tearDown(self):
        self.marionette.instance.app_args = self.orig_arguments
        self.marionette.quit(in_app=False, clean=True)

        super(TestCommandLineArguments, self).tearDown()

    def test_debugger_address_cdp_status(self):
        # By default Remote Agent is not enabled
        debugger_address = self.marionette.session_capabilities.get(
            "moz:debuggerAddress"
        )
        self.assertIsNone(debugger_address)

        # With BiDi only enabled the capability shouldn't be returned
        self.marionette.set_pref("remote.active-protocols", 1)
        self.marionette.quit()

        self.marionette.instance.app_args.append("-remote-debugging-port")
        self.marionette.start_session()

        debugger_address = self.marionette.session_capabilities.get(
            "moz:debuggerAddress"
        )
        self.assertIsNone(debugger_address)

        # Clean the profile so that the preference is definetely reset.
        self.marionette.quit(in_app=False, clean=True)

        # With all protocols enabled again the capability has to be returned
        self.marionette.start_session()
        debugger_address = self.marionette.session_capabilities.get(
            "moz:debuggerAddress"
        )

        self.assertEqual(debugger_address, "127.0.0.1:9222")
        result = requests.get(url="http://{}/json/version".format(debugger_address))
        self.assertTrue(result.ok)

    def test_websocket_url(self):
        # By default Remote Agent is not enabled
        self.assertNotIn("webSocketUrl", self.marionette.session_capabilities)

        # With CDP only enabled the capability is still not returned
        self.marionette.set_pref("remote.active-protocols", 2)

        self.marionette.quit()
        self.marionette.instance.app_args.append("-remote-debugging-port")
        self.marionette.start_session({"webSocketUrl": True})

        self.assertNotIn("webSocketUrl", self.marionette.session_capabilities)

        # Clean the profile so that the preference is definetely reset.
        self.marionette.quit(in_app=False, clean=True)

        # With all protocols enabled again the capability has to be returned
        self.marionette.start_session({"webSocketUrl": True})

        session_id = self.marionette.session_id
        websocket_url = self.marionette.session_capabilities.get("webSocketUrl")

        self.assertEqual(
            websocket_url, "ws://127.0.0.1:9222/session/{}".format(session_id)
        )

    # An issue in the command line argument handling lead to open Firefox on
    # random URLs when remote-debugging-port is set to an explicit value, on macos.
    # See Bug 1724251.
    def test_start_page_about_blank(self):
        self.marionette.quit()
        self.marionette.instance.app_args.append("-remote-debugging-port=0")
        self.marionette.start_session({"webSocketUrl": True})
        self.assertEqual(self.marionette.get_url(), "about:blank")

    def test_startup_timeout(self):
        try:
            self.marionette.quit()
            with self.assertRaisesRegexp(IOError, "Process killed after 0s"):
                # Use a small enough timeout which should always cause an IOError
                self.marionette.start_session(timeout=0)
        finally:
            self.marionette.start_session()
